# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '0.0.34'
version: str = '0.0.34'
full_version: str = '0.0.34'
git_revision: str = '2e37c151e3bd7e30470e084de4507ab46980f45d'
release: bool = True
if not release:
    version = full_version
